--
-- Pickup
-- Specialisation for lift and lower the pickup, ground ref, pressure pad and block the pickup if its overloaded
--
-- @author  Stefan Maurus	
-- @date  04/06/14

-- Copyright © Stefan Maurus, www.stefanmaurus.de

-- Includes script from fruktor to block the pickup (source: ClaasQuadrant1200.lua)

--[[
	<pickup>
		<animation name="ANIMATIONNAME" rake="INDEX" groundRef="INDEX" /> <!--groundRef has to be on the buttom of the pickup-->
		<pressurePad index="INDEX" minRot="DEGREES" maxRot="DEGREES"/>
		<dirt wheat="INDEX" grass="INDEX" cleanNode="INDEX"/> <!--You have to be near cleanNode to clean the pickup -->
		<blockSound file="PATH" volume="1"/>
		<grassCleaningParticle node="INDEX" position="0 0 0" rotation="0 0 0" file="PATH"/>
		<wheatCleaningParticle node="INDEX" position="0 0 0" rotation="0 0 0" file="PATH"/>
	</pickup>
]]	
	
Pickup = {};

function Pickup.prerequisitesPresent(specializations)
    return true;
end;

function Pickup:load(xmlFile)	
	self.getIsLoadingFromAir = Pickup.getIsLoadingFromAir;
	--#
	self.maxSpeedLevel = 2;
	self.wasToFast = false;
	self.speedViolationTimer = 0;
	self.speedViolationMaxTime = 500;
	self.setIsBlocked = SpecializationUtil.callSpecializationsFunction("setIsBlocked");
	self.isBlocked = false;
	self.blockTimer = 0;
	self.blockMaxTime = 1500;	
	local name = getXMLString(xmlFile, "vehicle.pickup.blockSound#file");
    self.blockSound = createSample("blockSound");
    loadSample(self.blockSound, Utils.getFilename(name, self.baseDirectory), false);		
	self.blockSoundVolume = getXMLFloat(xmlFile, "vehicle.pickup.blockSound#volume");
	
	self.maxLoad = getXMLFloat(xmlFile, "vehicle.pickup#maxload");
	
	--#
	self.isFirstRun = true;
	self.lastDeltaFillMean = 0;
	self.lastDeltaFill = 0;
	self.lastDeltaFillRun = 0;
	self.lastDeltaFillRunMax = 24;	-- ? 24 ~ 1 sec @ 1/40ms
	self.lastDeltaFillArray = {0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0};	-- 24 ?
	self.isFirstRun2 = true;
	self.lastDeltaFillMean2 = 0;
	self.lastDeltaFill2 = 0;
	self.lastDeltaFillRun2 = 0;
	self.lastDeltaFillRunMax2 = 24;	-- ? 24 ~ 1 sec @ 1/40ms
	self.lastDeltaFillArray2 = {0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0, 0,0,0};	-- 24 ?
	self.lastFillChangeTime = 0;
	
	self.lastFillDeltas = {};
	
	self.collectLastFillLevel = 0;
	self.collectLastFillLevel2 = 0;
	
	--#
	self.actLoad = 0;
	self.actLoad_Int = 0;
	self.actLoad_IntSend = 0;
	self.setActLoad = SpecializationUtil.callSpecializationsFunction("setActLoad");
	
	self.testtest = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.fillVolumeHeights.fillVolumeHeight.node#index"));

	--# 
	self.pudirt = {};
	self.pudirt.grass = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.pickup.dirt#grass"));
	self.pudirt.wheat = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.pickup.dirt#wheat"));
	self.pucleanNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.pickup.dirt#cleanNode"));
	self.cleaningEndTime = 0;
	setVisibility( self.pudirt.grass, false );
	setVisibility( self.pudirt.wheat, false );
	
	local psPath = string.format("vehicle.pickup.grassCleaningParticle");
	local cleanPSGrass = {};
	local particleNode = Utils.loadParticleSystem(xmlFile, cleanPSGrass, psPath, self.components, false, "$data/vehicles/particleSystems/trailerDischargeParticleSystem.i3d", self.baseDirectory);
	self.cleanPSGrass = cleanPSGrass;	

	psPath = string.format("vehicle.pickup.wheatCleaningParticle");
	local cleanPSWheat = {};
	local particleNode = Utils.loadParticleSystem(xmlFile, cleanPSWheat, psPath, self.components, false, "$data/vehicles/particleSystems/trailerDischargeParticleSystem.i3d", self.baseDirectory);
	self.cleanPSWheat = cleanPSWheat;	
	
	self.hud2 = {};
	local width, height = getNormalizedScreenValues(333, 30);
	local w, h = getNormalizedScreenValues(0, 244);	
	self.hud2.xPos = 1-width;
	self.hud2.yPos = h;
	self.hud2.hudWidth = width;	
	self.hud2.overlay_hud 	= Overlay:new("loadHUD_hud", 	Utils.getFilename("hud/hud_block.dds", self.baseDirectory), self.hud2.xPos, self.hud2.yPos, width, height);
	self.hud2.overlay_bar 	= Overlay:new("loadHUD_bar", 	Utils.getFilename("hud/hud_barY.dds", self.baseDirectory), self.hud2.xPos, self.hud2.yPos, width, height);	

	self.getIsAreaActive = Utils.overwrittenFunction(self.getIsAreaActive, Pickup.getIsAreaActive);	

	self.pickup = {};
	self.pickup.pressurePad = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.pickup.pressurePad#index"));
	self.pickup.pressurePadMin = math.rad(getXMLFloat(xmlFile, "vehicle.pickup.pressurePad#minRot"));
	self.pickup.pressurePadMax = math.rad(getXMLFloat(xmlFile, "vehicle.pickup.pressurePad#maxRot"));

	local name = getXMLString(xmlFile, "vehicle.pickup.lowerSound#file");
    self.lowerSound = createSample("lowerSound");
    loadSample(self.lowerSound, Utils.getFilename(name, self.baseDirectory), false);		
	self.lowerSoundVolume = getXMLFloat(xmlFile, "vehicle.pickup.lowerSound#volume");	
	
    self.setVehicleIncreaseRpm = SpecializationUtil.callSpecializationsFunction("setVehicleIncreaseRpm");
    self.saveMinimumRpm = 0;

	self.cTime = 0;
	
	self.pickUpSoundPlayed = false;
end;


function Pickup:postLoad(xmlFile)
end;


function Pickup:delete()
	Utils.deleteParticleSystem(self.cleanPSGrass);
	Utils.deleteParticleSystem(self.cleanPSWheat);
end;

function Pickup:readStream(streamId, connection)
	self.isBlocked = streamReadBool(streamId);
	self:setActLoad(streamReadFloat32(streamId), true);
end;

function Pickup:writeStream(streamId, connection)	
	streamWriteBool(streamId, self.isBlocked);
	streamWriteFloat32(streamId, self.actLoad_IntSend);
end;

function Pickup:getIsAreaActive(superFunc, area)
    if superFunc ~= nil then
        return superFunc(self, area) and self.movingDirection > 0 and self.isPickupLowered; --just load when pickup is down and moving forward
    end;
	return self.movingDirection > 0 and self.isPickupLowered;
end;

function Pickup:readUpdateStream(streamId, connection, dirtyMask)
end;

function Pickup:writeUpdateStream(streamId, connection, dirtyMask)
end;

function Pickup:mouseEvent(posX, posY, isDown, isUp, button)
end;

function Pickup:keyEvent(unicode, sym, modifier, isDown)
end;

function Pickup:update(dt)
	if self.isClient then
		if self:getIsActiveForSound() then
			if self.isPickupLowered and not self.pickUpSoundPlayed then
				self.pickUpSoundPlayed = true;
				playSample(self.lowerSound, 1, self.lowerSoundVolume, 0);
			end;	
			
			if not self.isPickupLowered and self.pickUpSoundPlayed then
				playSample(self.lowerSound, 1, self.lowerSoundVolume, 0);
				self.pickUpSoundPlayed = false;
			end;
		end;
	end;
	
	if self.cleaningEndTime > 0 then
		self.cleaningEndTime = math.max(self.cleaningEndTime-dt, 0);
	end;
	
	self.cTime = self.cTime + dt;
	
	if self.isBlocked and self.isTurnedOn then
		self:setIsTurnedOn(false);
	end;
	
	--#
	if g_currentMission.player ~= nil then
		local pvx, pvy, pvz = getWorldTranslation(g_currentMission.player.rootNode);
		local px, py, pz = getWorldTranslation(self.pucleanNode);
		local distance = Utils.vector3Length(px-pvx, py-pvy, pz-pvz);
		if distance < 2 then
			if self.isBlocked then
				g_currentMission:addHelpButtonText( g_i18n:getText("PICKUP_CLEAN"), InputBinding.IMPLEMENT_EXTRA);
				if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA) then
					self:setIsBlocked(false);
				end;
			end;
		end;
	end;	

	if self.actLoad_IntSend > 0 and not self.isBlocked then
		local _,y,z = getRotation(self.pickup.pressurePad);
		local minMax = self.pickup.pressurePadMax-self.pickup.pressurePadMin;
		local minMaxPer = minMax*self.actLoad_IntSend;	
		setRotation(self.pickup.pressurePad, (self.pickup.pressurePadMin+minMaxPer),y,z);
	elseif self.isBlocked then
		local _,y,z = getRotation(self.pickup.pressurePad);
		setRotation(self.pickup.pressurePad, self.pickup.pressurePadMax,y,z);	
	else
		local _,y,z = getRotation(self.pickup.pressurePad);
		setRotation(self.pickup.pressurePad, self.pickup.pressurePadMin,y,z);	
	end;	
end;

function Pickup:updateTick(dt)
	Cylindered.updateMovingPart(self, self.activeDirtyMovingParts[1], true);
	Cylindered.updateMovingPart(self, self.activeDirtyMovingParts[2], true);
	Cylindered.updateMovingPart(self, self.activeDirtyMovingParts[3], true);
	Cylindered.updateMovingPart(self, self.activeDirtyMovingParts[4], true);
		
	--# speed control
	-- for client
	local deltaFill = 0;
	if self.isClient then
		if self.collectLastFillLevel ~= self.fillLevel then
			if self.fillLevel > self.collectLastFillLevel then
				deltaFill = self.fillLevel - self.collectLastFillLevel;
			else
				deltaFill = self.lastDeltaFill;
			end;			
			self.collectLastFillLevel = self.fillLevel;
			self.lastFillChangeTime = self.cTime;
		end	
	end;	
	
	--# host
	if self.isServer then
		local deltaFill2 = 0;
		if self.collectLastFillLevel2 ~= self.fillLevel then
			if self.fillLevel > self.collectLastFillLevel2 then
				deltaFill2 = self.fillLevel - self.collectLastFillLevel2;
			else
				deltaFill2 = 0; --self.lastDeltaFill2;
			end;			
			self.collectLastFillLevel2 = self.fillLevel;
		end	

		if deltaFill2 ~= 0 and self.isFirstRun2 == false then
			table.insert(self.lastFillDeltas, {t=self.cTime, d=deltaFill2});
		end;
		if self.isFirstRun2 == true then
			self.isFirstRun2 = false;
		end;
		self.lastDeltaFill2 = deltaFill2;
		
		local fill = 0;
		if table.getn(self.lastFillDeltas) > 0 then
			local noMoreOld = false;
			while not noMoreOld do
				if table.getn(self.lastFillDeltas) > 0 then
					for i,lfd in pairs(self.lastFillDeltas) do
						if self.cTime - lfd.t > 1000 then	
							table.remove(self.lastFillDeltas, i);
							break;
						end
						if i >= table.getn(self.lastFillDeltas) then
							noMoreOld = true;
						end;
					end;
				else
					break;
				end;
			end
			for i,lfd in pairs(self.lastFillDeltas) do
				fill = fill + lfd.d;
			end;
		end
		self:setVehicleIncreaseRpm(dt, fill*3, true);	
		--DURAL : increasead max load
		local maxLoad = self.maxLoad; -- 150
		self.actLoad2 = math.max(0, math.min(1, fill / maxLoad));
		if self:getIsLoadingFromAir() then
			self.actLoad2 = 0;
		end;
		self:setActLoad( self.actLoad2 ); --math.floor( self.actLoad2*255 + 0.5 ) );

		if self.actLoad2 == 1 and self.isServer then --self.lastDeltaFillMean > maxLoad then		
			self.blockTimer = self.blockTimer + dt;
			if self.blockTimer > self.blockMaxTime then
				if not self.isBlocked then
					self.blockTimer = 0;
					self:setIsBlocked(true);
					self.isFirstRun2 = true;
				end
			end;
		elseif self.isServer then
			self.blockTimer = math.max(0, self.blockTimer - dt/10);
		end;
		
	end;	
	--#
	if self.isClient then
		if self.cleaningEndTime > 0 then
			if self.currentFillType == Fillable.fillTypeNameToInt["grass_windrow"] then
				Utils.setEmittingState(self.cleanPSGrass, true);
			else
				Utils.setEmittingState(self.cleanPSWheat, true);	
			end;
		else
			Utils.setEmittingState(self.cleanPSGrass, false);
			Utils.setEmittingState(self.cleanPSWheat, false);	
		end;	
	end;
	
end;

function Pickup:draw()
	
	setTextColor(1.0, 1.0, 1.0, 1.0);
	
	--DURAL : don't forget to set the alignment and bold !
	setTextBold(false);			
	setTextAlignment(RenderText.ALIGN_LEFT);
	
	--# actual load hud

	local percentage = self.actLoad_IntSend; --/255; --self.actLoad;
	
	if percentage > 0.95 then
		setOverlayColor(self.hud2.overlay_bar.overlayId, 1, 0, 0, 1)	
	else
		setOverlayColor(self.hud2.overlay_bar.overlayId, 0, 1, 0, 1)
	end;
	
	self.hud2.overlay_bar.width = self.hud2.hudWidth * percentage;	
	setOverlayUVs(self.hud2.overlay_bar.overlayId, 0, 0.05, 0, 1, percentage, 0.05, percentage, 1);
	self.hud2.overlay_bar:render();	
	
	self.hud2.overlay_hud:render();	
	
	setTextAlignment(RenderText.ALIGN_CENTER);
	local x, y = getNormalizedScreenValues(333/2, 10.5);
	renderText(self.hud2.xPos+x, self.hud2.yPos+y, 0.012, g_i18n:getText("ACT_LOAD").." "..string.format("%5.2f %%", 100*self.actLoad_IntSend)); --/255)); --100*self.actLoad)); -- 
	setTextAlignment(RenderText.ALIGN_LEFT);

	if self.isBlocked then
		g_currentMission:showBlinkingWarning(g_i18n:getText("PICKUP_IS_BLOCKED"), 10000);
	end;
end;

function Pickup:getIsLoadingFromAir()
	if self.isTurnedOn or self.topRopesAreVis then
		return false;
	end;
	return true;
end;

--#
function Pickup:setIsBlocked(state, noEventSend)
	SetBlockedEvent.sendEvent(self, state, noEventSend);
	self.isBlocked = state;
	if state == true then
		if self.isClient and self:getIsActiveForSound() then
			playSample(self.blockSound, 1, self.blockSoundVolume, 0);
		end;		
		--if self.
		if self.currentFillType == Fillable.fillTypeNameToInt["grass_windrow"] then
			setVisibility( self.pudirt.grass, true);
		else	
			setVisibility( self.pudirt.wheat, true);
		end;
	else
		setVisibility( self.pudirt.grass, false);
		setVisibility( self.pudirt.wheat, false);
		self.cleaningEndTime = 300;
	end	
end;

function Pickup:setActLoad(load, noEventSend)
	SetActLoadEvent.sendEvent(self, load, noEventSend);
	self.actLoad_IntSend = load;
end;

function Pickup:setVehicleIncreaseRpm(dt, increase, isActive)
	if self.attacherVehicle ~= nil and self.saveMinimumRpm ~= 0 and
		self.attacherVehicle.motor ~= nil then
		if dt ~= nil then
			if isActive == true then
				self.attacherVehicle.motor.minRpm = math.max(self.attacherVehicle.motor.minRpm-(dt*2), -increase);
			else
				self.attacherVehicle.motor.minRpm = math.min(self.attacherVehicle.motor.minRpm+(dt*5), self.saveMinimumRpm);
			end;
		else
			self.attacherVehicle.motor.minRpm = self.saveMinimumRpm;
		end;
		if self.attacherVehicle.isMotorStarted then
			local fuelUsed = 0.0000012*math.abs(self.attacherVehicle.motor.minRpm);
			self.attacherVehicle:setFuelFillLevel(self.attacherVehicle.fuelFillLevel-fuelUsed);
			g_currentMission.missionStats.fuelUsageTotal = g_currentMission.missionStats.fuelUsageTotal + fuelUsed;
			g_currentMission.missionStats.fuelUsageSession = g_currentMission.missionStats.fuelUsageSession + fuelUsed;
		end;
	end;
end;

--

SetActLoadEvent = {};
SetActLoadEvent_mt = Class(SetActLoadEvent, Event);

InitEventClass(SetActLoadEvent, "SetActLoadEvent");

function SetActLoadEvent:emptyNew()
    local self = Event:new(SetActLoadEvent_mt);
    self.className="SetActLoadEvent";
    return self;
end;

function SetActLoadEvent:new(vehicle, actLoad)
    local self = SetActLoadEvent:emptyNew()
    self.vehicle = vehicle;
	self.actLoad = actLoad;
	return self;
end;

function SetActLoadEvent:readStream(streamId, connection)
    local id = streamReadInt32(streamId);
    self.vehicle = networkGetObject(id);
	self.actLoad = streamReadFloat32(streamId);
    self:run(connection);
end;

function SetActLoadEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.vehicle));	
	streamWriteFloat32(streamId, self.actLoad);
end;

function SetActLoadEvent:run(connection)
	self.vehicle:setActLoad(self.actLoad, true);
	if not connection:getIsServer() then
		g_server:broadcastEvent(SetActLoadEvent:new(self.vehicle, self.actLoad), nil, connection, self.vehicle);
	end;
end;

function SetActLoadEvent.sendEvent(vehicle, actLoad, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(SetActLoadEvent:new(vehicle, actLoad), nil, nil, vehicle);
		else
			g_client:getServerConnection():sendEvent(SetActLoadEvent:new(vehicle, actLoad));
		end;
	end;
end;

SetBlockedEvent = {};
SetBlockedEvent_mt = Class(SetBlockedEvent, Event);

InitEventClass(SetBlockedEvent, "SetBlockedEvent");

function SetBlockedEvent:emptyNew()
    local self = Event:new(SetBlockedEvent_mt);
    self.className="SetBlockedEvent";
    return self;
end;

function SetBlockedEvent:new(vehicle, state)
    local self = SetBlockedEvent:emptyNew()
    self.vehicle = vehicle;
	self.state = state;
	return self;
end;

function SetBlockedEvent:readStream(streamId, connection)
    local id = streamReadInt32(streamId);
	self.state = streamReadBool(streamId);
    self.vehicle = networkGetObject(id);
    self:run(connection);
end;

function SetBlockedEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.vehicle));	
	streamWriteBool(streamId, self.state);
end;

function SetBlockedEvent:run(connection)
	self.vehicle:setIsBlocked(self.state, true);
	if not connection:getIsServer() then
		g_server:broadcastEvent(SetBlockedEvent:new(self.vehicle, self.state), nil, connection, self.object);
	end;
end;

function SetBlockedEvent.sendEvent(vehicle, state, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(SetBlockedEvent:new(vehicle, state), nil, nil, vehicle);
		else
			g_client:getServerConnection():sendEvent(SetBlockedEvent:new(vehicle, state));
		end;
	end;
end;